#!/usr/local/bin/perl
#
# OMICHAIN
#
# XMM Pipeline Processing System OMICHAIN script
#
# Author: Jon Rainger (jfr@mssl.ucl.ac.uk)
#
# Description: This script runs the tasks required for to produce
#              PPS products from Imaging Mode ODFs.
#
# $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
#
#
#    MODIFICATIONS:
#    12/06/2000 - AMK - Removed call to OMPSF as none of the pipleline tasks use it's output.
#
#    12/06/2000 - AMK - Redirected sreen output to a file omichain_log as user will lose the output during a complete ODF
#
#    06/09/2000 - AMK - Ammended code so that the new style house keeping files are processed correctly.
#
#    06/09/2000 - AMK - Ammended code so that product names conform to specification.
#
#    29/09/2000 - AMk - Changed flat field product file name to tmp_flat_field as this file should NOT be a product.
#
#    11/10/2000 - AMK - Removed the die calls and replace with SAS::fatal calls.
#
#    03/11/2000 - CB  - Modified so that images with "BLOCKED" or "UNKNOWN" filters are not processed.
#                       Added call to ps2pdf.
#    21/01/2000 - CB - Modified so that "ommosaic" runs at the end to produce a composite image for each filter.
#    10/08/2001 - CB - Modified to process engineering 2 and 4 data and restructured.
#    11/20/2001 - CB - Modified to accept command-line parameters for ommodmap, omdetect, ommodmap, ommosaic
#                      and omsrclistcomb.     
#    12/12/2001 - CB - Modified to process the exposures for each filter in the order that the filters 
#                      appear in the SAS summary file.                       
#    08/-5/2002 - CB - Modified to keep the intermediate source-list files from omdetect & ommag
#    03/12/2002 - CB - Modified to use the omatt parameter rotateimage
#    15/06/2003 - CB - Perl script overhauled for latest Perl
#    27/01/2004 - VNY - an error with input parameters and verbosity > 7 is by-passed
#    01.02.2004 - VNY the verbosity level is lowered to 5 for the moment 
#                 of reading the input parameters of omichain
#    14.04.2004 - VNY - protected against using an uninitialized string 
#                       variable in the lines 398-413 (to fix SSC-SPR-3286)
# 29.05.2004 VNY   Correction in the subroutine GetODFDirectory to protect 
#                  the task from failure during its test in Mac-machines 
#************************************************************************************************************************
#use DAL;
use SAS;
my @FilterList =();                   # Stores a list of the filters in the summary file
my @FilterFileList = (); 
my @ExposureNoList =();               # Stores a list of the exposure numbers
my %ExposureFilterList =();           # Stores a list of the filter for each exposure number
my @E2IMIList =();                    # Stores the IMI files created by omengine2
my @E4IMIList =();                    # Stores the engineering 4 mode image files
my @Simage2000Files = ();   
my %ExposureType = ();
my $SASFile = "TEST";
my @FileList = (); 

my @InputFilterList = ();    
my @OMFilterCodes = ();
my @ExposureImageList = ();
my %ExposureExtension = ();
my %ExposureScheduleList = ();

my (@UserExposureList) = @_;         # The list of exposures in the parameter list
my (@OMFilterList); # The list of OM filters from the SAS summary file
   
my (@ExposureList) = ();
my @osw_file_list = ();

my %orb_list = ();
 


my $inp_directory = "";  # The input directory
my $out_directory = "";  # The output directory
my $redirect = 0; 

#*************************************************
# Output files
#************************************************
my $tracking_history_filename;
my $in_orbit_flat_field_filename;
my $flatFieldFileName;
my $detectorCoordImageFileName;
my $IntermediateImage1FileName;
my $IntermediateImage2FileName;
my $modulo_8_product_filename;
my $regionFileName;
my $levelImageFileName;
my $backgroundImageFileName;

my $osw_list_intermediary_detect_filename;
my $osw_list_intermediary_detect_filename1;
my $osw_list_intermediary_detect_filename2;
my $trackingHistoryFileName;
my $tracking_history_plot_filename;
my $tracking_history_plot_filename1;
my  $window_data_filename; 
my $periodic_hk_filename;
my $non_periodic_hk_filename;
my $tracking_history_timeseries_filename;
my $observation_source_list_pps_product_filename;
my $skyCoordImageFileName;
my $flat_field_product_filename;


my $periodic_file_flag;
my $non_periodic_file_flag;

my $obs_key;
my $win_key;
my $exp_type;
my $orb_key;

#*********************************************
# Input parameters
#*********************************************
my $ommodmap_nsig;
my $ommodmap_nbox;
my $omdetect_nsigma;
my $omdetect_minsignificance;
my $omdetect_detectextended;
my $omsrclistcomb_nsigma;
my $omsrclistcomb_usecat;
my $maxradecerr;
my $maxrmsres;
my $ommosaic_exposure;
my $ommosaic_exposuremap;
my $omatt_usecat;
my $catfile;
my $omatt_rotateimage = 1;




sub omichain 

{


   

#/////////////////////////////////////////
    my $SASVERBOSITY = " ";
    if (exists($ENV{SAS_VERBOSITY}))
    {
	$SASVERBOSITY = $ENV{SAS_VERBOSITY};
    }
    else
    {
	$SASVERBOSITY = "5";
	$ENV{SAS_VERBOSITY} = "5";
    }
    if ($SASVERBOSITY > 7) 
    {
	$ENV{SAS_VERBOSITY} = "5";
    }

    #************************************************
    # For program development message can be written
    # to a file "omichain_log"
    # REMEMBER TO BLANK OUT NEXT LINE BEFORE UPLOADING
    #***********************************************
    $redirect = 1;       
#    &CreateLogFile;

   
  

    #**************************************
    # Set up the directory paths for input
    # and output.
    #**************************************
    &SetUpDirectoryPaths;

  
    my $taskname='OMICHAIN';
      
   
    my $VERSION='1.48';
    my $date = &date_time;
    my $comment = stringParameter("comment");
########## VNY 27.01.2004
    my $commentLength=length($comment);
    if ($commentLength > 255)
    {
	$comment=" ";
	$commentLength=1;
 
   }

 
 

##########  
    my $create = "Running SAS task $taskname V$VERSION $date";      
    my $text1 = "Input directory = $inp_directory";  
    my $text2 = "Output directory = $out_directory"; 
      
    &HighlightedMessage("*", $create, $comment, "   ", $text1, $text2);
      
    #**************************************
    # Get the parameters for ommodmap,
    # omdetect, omatt, omsrclistcomb & ommosaic
    #**************************************
    &GetOmmodmapParameters;

    &GetOmdetectParameters;
 
    &GetOmattParameters;

 
    &GetOmsrclistcombParameters;


    &GetOmmosaicParameters;


    #**************************************
    # Get the list of filters that the
    # user wants to process (if any in list)
    #**************************************
  
    @InputFilterList = &GetStringParameterList("filters");


# VNY 01.02.2004 Return SAS_VERBOSITY to its requested level
    if ($SASVERBOSITY > 7) 
    {
	$ENV{SAS_VERBOSITY} = $SASVERBOSITY;
    }
       
    #*****************************************************
    # Read the SAS summary file to get a list of 
    # observations, etc, to be processed
    #***************************************************** 
    &ExamineSASFile;     
  

    
    #***************************************************
    # Initialise the lists of engineering 2/4 data, etc.
    #***************************************************

    &InitialiseLists;
   
    #************************************************** 
    # Check for the presence of the house keeping files
    #**************************************************
    &CheckForHouseKeepingFiles;

    #**********************************************
    # Run the pipeline- Loop over all observations.
    #**********************************************
   
    foreach $orb_key (keys(%orb_list))
    {


        #********************************************
        # Process the data for each orbit in the list
        #********************************************
        &ProcessOrbit($orb_key, @InputFilterList);  

    }
  
    #********************************************
    # We don't want to keep any new image file
    # created from Engineering mode 2 data
    #********************************************
#    &DeleteNewImageFiles;

    $date = &date_time;
       
    my $message = "Finished running SAS task $taskname V$VERSION $date";
    &HighlightedMessage("*", $message);

    

    #******************************
    # For test purposes
    #******************************
    if ($redirect == 5)
    {
        close(STDOUT);
        close(STDERR);
    }
   
}
#***********************************************************
#
# Subroutine section
#
#***********************************************************
sub date_time {          # create the file creation datestring

    my  $datestring;
    $datestring = `date -u`;
    chop $datestring;
    return $datestring;
}
#**************************************
# subroutine GetFilterCode
# Given a filter from the SAS file
# returns the filter code to be used
# in constructing the output file name.
#**************************************
sub GetFilterCode
{
    my($filter) = @_;  # The input filter name
    
    
    my @OMFilters = ("BLOCKED", "V",  "Magnifier", "MAGNIFIER", "U",  "B",  "WHITE", "GRISM2", "Grism2", "UVW1", "UVM2", "UVW2", "GRISM1", "Grism1");

    my @OMFilterCodes = ("BL",  "V", "MA", "MA", "U", "B", "W", "G",  "G",  "L",   "M",   "S", "H", "H");


    for (my $i = 0; $i < @OMFilters; $i++)
    {
       
        if ($OMFilters[$i] eq $filter)
        {
            return $OMFilterCodes[$i];
        
	}
    }
    die("No such filter ($filter) in filter table!");
    
}
#************************************
# Sub IsHighResCentralWindow
#************************************
sub IsHighResCentralWindow
{
    my ($fileName) = @_;
                 
    if($fileName =~ /OMS00/)
    {
       
        if($fileName =~ /IMAGE_1000/)
        {
            return 1;
       
        }     
    }    
    return 0;
}
#********************************************************************
# subroutine FindEngineering4ImageFiles
#********************************************************************
sub FindEngineering4ImageFiles
{
    my $numExposures = @ExposureNoList;
  
    my @ImageFile = <$inp_directory/*E4I.FIT*>;        

#    print("Findengineering4Imagefiles line 356\n");

    my $numFiles = @ImageFile;
   
    if($numFiles == 0) 
    {
        return;   
    }


#    print("Findengineering4imagefiles line 363\n");


    #************************************************************************
    # Create a list of engineering 4 image files, without the directory path
    #************************************************************************
    for (my $i = 0; $i < $numFiles; $i++)
    {
        my $fileName = $ImageFile[$i];       
        my $pos = index($fileName, "E4I") - 24;
        my $name = substr($fileName, $pos, 32);    # Get rid of the directory name                        



#	print("Line 377 $i) file=$fileName\n");

        push(@E4IMIList, $name);                          # Add it to the list of engineering mode 4 data
      

        $fileName = $name;

        #*********************************************
        # Find out which exposure this file is
        #*********************************************
        my $exposure = substr($fileName, 19, 3);      # The exposure number
    
        my $found = 0;
        my $j = 0;

        while ($j < $numExposures and $found == 0)
        {
            
            my $exposureNo = substr($ExposureNoList[$j], 0, 3);
        
            if ($exposure eq  $exposureNo)
            { 
                $ExposureType{$exposureNo} = "ENG4";
                $ExposureExtension{$exposureNo} = "*E4I.FIT*";
	    
		$found = 1;
                       
            }
            $j++;
        }

    }
}
#******************************************************
# Subroutine FillExposureList
# Obtains a list of the exposures in the summary file
#******************************************************  
sub FillExposureNoList
{
   
    open (SASFILE, $SASFile) || die("Unable to open SAS SUMMARY FILE $SASFile- Please check directory paths and SAS_ODF variable"); 

    my $line;
   
#  print("fillexposurenolist line 414\n");

   
    #*************************************************
    # Skip past the EPIC and RGS stuff
    #*************************************************
 
    while ($line = <SASFILE>) {
	my $k=substr($line, 0, 2);
	next unless $k eq "OM";
	last;
    }
       
#    until (substr($line, 0, 2) eq "OM")
#    {
#
#       $line = <SASFILE>;
#                                  
#    } 

    while ($line = <SASFILE>)
    {
       
        if ($line =~ "/ Exposure Identifier")
        {        
 
	  
            my $exposureNo = substr($line, 0, 3);
            my $exposureSchedule = substr($line, 32, 1); 
            push(@ExposureNoList, $exposureNo);      
                 
            $ExposureScheduleList{$exposureNo} = $exposureSchedule;
            #****************************************
            # Get the filter name from the SAS file
            #****************************************
	  
	     my $filterName = &GetNextSASFilter;
	   
            if ($filterName eq "")
            {
                last;
            }
            #************************************************************************
            #  Store the exposure number for this filter
            #************************************************************************
           
            %ExposureFilterList = Store($filterName, $exposureNo, %ExposureFilterList);

	    
	       
        }
    }

    
#    close $handle;
    close SASFILE;
    
 

}
#***************************************************************
# Subroutine DeleteNewImageFiles
# Renames the new image files created from each set of four
# engineering 2 mode image files.
#***************************************************************
sub DeleteNewImageFiles
{

    if ($out_directory ne $inp_directory)
    { 
        my $num = @E2IMIList;   


        for (my $j = 0; $j < $num; $j++)
        {
        
            my $fileName = $E2IMIList[$j];       
            my $pos = index($fileName, "E2I") - 24;
            my $name = $out_directory . '/' . substr($fileName, $pos, 32);    # Get rid of the directory name               
    
      
            system("cp $fileName $name"); 
        }         
        	    		
        unlink (@E2IMIList);
    }
              


    #*********************************************
    # Move the flatfield file created by omflatgen
    # to the output directory and delete the one
    # in the input directory
    #*********************************************
    if ($flatFieldFileName ne "BLANK" and $out_directory ne $inp_directory) 
    {
        my $name = $inp_directory . '/'. $flatFieldFileName;
        system("cp $flatFieldFileName  $name"); 
        unlink $flatFieldFileName;
    }

    #******************************************
    # Delete temp files
    #*****************************************
    chdir $out_directory;

    &DeleteFileList("tmp");

    #******************************************
    # Delete imi files
    #*****************************************
    &DeleteFileList("imi");
 
    #******************************************
    # Delete fit files
    #*****************************************
    &DeleteFileList("fit");
 

}
#***************************************
# DeleteFileList
#***************************************
sub DeleteFileList($)
{ 
    my ($name) = @_;
   
    my @List = <*$name*> ;    # Creates a list of FIT file names    

    unlink (@List);
}

#**************************************************************
# Subroutine MissingImageFileWarning
#*************************************************************
sub MissingImageFileWarning($)
{ 


#    my ($fileName) = @_;
    my $exp_key = " ";
    my $value = $ExposureType{$exp_key};
    
    if ($value eq "ENG2" or $value eq "ENG4" and $value ne "")
    {
        return;
    }

    my $imi_file = $orb_key . "_" . $obs_key . "_OM" . $exp_type . $exp_key . $win_key . "IMI.FIT";
    
	
    &Message("*** Warning *************************************************");

    my $message = "Warning! - Window '$win_key' of exposure '$exp_key' in observation '$obs_key' is missing the image file $imi_file.";
    &HighlightedMessage("*", $message);
   
  
    
}
#********************************************
# Subroutine CreatePoductFileNames
#********************************************
sub CreateProductFileNames($)
{
    
    my ($observationNo, $exposureSchedule, $exposureNo, $windowNo, $filter) = @_;           # The exposure and window numbers
    $windowNo = substr($windowNo, 1, 1);
    my $exposureType = $ExposureType{$exposureNo}; # Rudi5, ENG2 or ENG4

     
    my $filterCode = &GetFilterCode($filter);
    my $string1;
    my $string2;
    my $string3; 
    if ($exposureType eq "RUDI5")       # Rudi-5 case
    {
        $string1 = "SIMAGE$windowNo";
        $string2 = "IMAGE_$windowNo";
        $string3 = $windowNo;
    }
    elsif ($exposureType eq "ENG2")     # Engineering-2 case
    {
        $string1 = "FSIMAG$filterCode";
        $string2 = "FIMAG_$filterCode";
        $string3 = $filterCode;
 
    }
    elsif ($exposureType eq "ENG4")     # Engineering-4 case
    {
        
        $string1 = "FSIMAG$filterCode";
        $string2 = "FIMAG_$filterCode";
        $string3 = $filterCode;
 
    }
    else
    {
        die("UNKNOWN EXPOSURE TYPE ($exposureType)");
    } 
    
    
    #***************************************************
    # The detector-coordinate (ie non-rotated image) file name
    #***************************************************
    $detectorCoordImageFileName = $out_directory . "/P$observationNo" . "OM$exposureSchedule$exposureNo$string2" . "000.FIT";
    

    #***************************************************
    # The sky-coordinate (ie rotated image) file name
    #***************************************************
    $skyCoordImageFileName = $out_directory . "/P$observationNo" . "OM$exposureSchedule$exposureNo$string1" . "000.FIT";

    

    #************************************
    # The intermediate image file names
    #***********************************
    $IntermediateImage1FileName = $out_directory . "/I$observationNo". "OM$exposureSchedule$exposureNo" . "IMAGE1$windowNo" . "000.FIT";
    $IntermediateImage2FileName = $out_directory . "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "IMAGE2$windowNo" . "000.FIT";
#    $IntermediateImage3FileName = $out_directory . "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "IMAGE3$windowNo" . "000.FIT";
    $flat_field_product_filename = $out_directory . "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "FLATF$windowNo" . "000.FIT";

    #******************************
    # Product created by ommodmap
    #*****************************
    $modulo_8_product_filename = $out_directory .  "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "MOD8$windowNo" . "000.FIT";

    #*****************************
    # Products created by omdetect
    #*****************************
    $regionFileName = $out_directory .  "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "REGION$windowNo" . "000.ASC";

    $levelImageFileName = $out_directory .  "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "LEVELIMAGE$windowNo" . "000.FIT";

    $backgroundImageFileName = $out_directory .  "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "BCKIMAGE$windowNo" . "000.FIT";






    #**************************************************
    # The source=list created by omdetect
    #**************************************************
    $osw_list_intermediary_detect_filename1 = "I$observationNo" . "OM$exposureSchedule$exposureNo"."SWSRLI1$string3"."000.FIT";


    #**************************************************
    # The source-list created by ommag
    #**************************************************
    $osw_list_intermediary_detect_filename2 = "I$observationNo" . "OM$exposureSchedule$exposureNo"."SWSRLI1$string3"."001.FIT";
    
    #**************************************************
    # The source-list created by ommat (the product file)
    #**************************************************
   
    
    $osw_list_intermediary_detect_filename = "P$observationNo" . "OM$exposureSchedule$exposureNo"."SWSRLI$string3"."000.FIT";



}
#********************************************
# Subroutine CheckTHXFileExits
#********************************************
sub CheckTHXFileExists($)
{
    
    my ($orbitNo, $observationNo, $exposureSchedule, $exposureNo) = @_;           # The exposure and window numbers
 

    
    $tracking_history_filename = $inp_directory . "/". $orbitNo . "_" . $observationNo . "_OM" . $exposureSchedule . $exposureNo . "00" . "THX.FIT";
   
  

    #***************************************************
    # Check for existence of file
    #**************************************************
    if (-e  $tracking_history_filename) 
    

    {
       
    }	 
    else    
    {
    
        #***********************************
        # Create a dummy name to fool OMPREP
        #***********************************
        $tracking_history_filename = $out_directory . "/DUMMYTHX.FIT";
	
 
    }

}

#************************************************
# Subroutine ProcessIageFile
# Processes an OM imaging file
#***********************************************
sub ProcessImageFile($)
{ 

    my ($image_filename, $observationNo, $exposureSchedule, $exposureNo, $windowNo, $filter) = @_;   

 
    my $exposureType = $ExposureType{$exposureNo};   # Rudi5, ENG2 or ENG4
   

    #******************************************
    # Set pipeline variables.
    #**************************
    &CreateProductFileNames($observationNo, $exposureSchedule, $exposureNo, $windowNo, $filter);


    #***********************************
    # Produce image mode products.
    #***********************************

    &HighlightedMessage("*", "Processing image file $image_filename");
    #*********************************
    # omprep
    #*********************************
    &HighlightedMessage("*", "omprep");
   

    my $arg_list = "set=$image_filename" .
                   " nphset=$non_periodic_hk_filename" .
                   " pehset=$periodic_hk_filename" .
                   " wdxset=$window_data_filename" .
		   " outset=$IntermediateImage1FileName";


#    &Message("omprep $arg_list");

    #**************************************************************
    # Modified so that if omprep fails it just skips the observation,
    # but does not kill the chain (1.11.2000)
    #**************************************************************
    system("omprep $arg_list");

    my $exit_value = $? >> 8;   # 0 okay, 1 call fatal called, or some other disaster
    if ($exit_value == 1) # Failure in omprep 
    { 
        &HighlightedMessage("*", "omprep has detected an error- this observation will not be processed");
        return;

    }     
    
    #******************************************
    # omcosflag
    #******************************************
    &HighlightedMessage("*", "omcosflag");

    $arg_list = "set=" .  $IntermediateImage1FileName .
                " thxset=" . $trackingHistoryFileName;

#    &Message("omcosflag $arg_list");
    system("omcosflag $arg_list") && SAS::fatal("Task Failure","omcosflag failed.");
 
  
    #****************************************
    # omflatfield
    #****************************************
    &HighlightedMessage("*", "omflatfield");

    $arg_list = "set=$IntermediateImage1FileName" .
                " thxset=$trackingHistoryFileName" .
                " inorbitflatset=$in_orbit_flat_field_filename" .
                " tsflatset=$flat_field_product_filename" .
                " outset=$IntermediateImage2FileName";

#    &Message("omflatfield $arg_list");

    system("omflatfield $arg_list") && SAS::fatal("Task Failure","omflatfield failed.");
               

    #***************************************
    # ommodmap
    #***************************************
    &HighlightedMessage("*", "ommodmap");

    $arg_list = "set=$IntermediateImage2FileName" .
                " mod8product=yes" .
		" mod8set=$modulo_8_product_filename" .
                " outset=$detectorCoordImageFileName" . 
#                " outset=$IntermediateImage3FileName" .
                " nsig=$ommodmap_nsig" .
                " nbox=$ommodmap_nbox";  

#    &Message("ommodmap $arg_list");
    system("ommodmap $arg_list") && SAS::fatal("Task Failure","ommodmap failed.");
                     


    #***********************************************
    # Stop processing image file if a grism image
    #***********************************************
  
   if ($filter eq "GRISM1" or $filter eq "GRISM2" or $filter eq "MAGNIFIER" or $filter eq "Magnifier") 
    {
        &HighlightedMessage("X", "Cannot process this image file any further ($filter image)");
        return 1;


    }
    #**************************************************
    # omdetect
    #**************************************************
    &HighlightedMessage("*", "omdetect");
    $arg_list = "nsigma=$omdetect_nsigma" .
                " minsignificance=$omdetect_minsignificance" .
                " detectextended=$omdetect_detectextended" .
                " set=$detectorCoordImageFileName" .
                " regionfile=$regionFileName" .
                " levelimage=$levelImageFileName" .
                " backgroundimage=$backgroundImageFileName" .
                " outset=$out_directory/$osw_list_intermediary_detect_filename";

 
#    &Message("omdetect $arg_list");

    system("omdetect $arg_list") && SAS::fatal("Task Failure","omdetect failed.");

    #*********************************************************************
    # Keep a copy of this source-list
    #*********************************************************************

    system("cp $out_directory/$osw_list_intermediary_detect_filename $out_directory/$osw_list_intermediary_detect_filename1");
    #*********************************************** 
    # ommag
    #***********************************************
    &HighlightedMessage("*", "ommag");

    $arg_list = "set=$out_directory/$osw_list_intermediary_detect_filename";
                

#    &Message("ommag $arg_list");

    system("ommag $arg_list") && SAS::fatal("Task Failure","ommag failed.");

    #*********************************************************************
    # Keep a copy of this source-list
    #*********************************************************************
    system("cp $out_directory/$osw_list_intermediary_detect_filename $out_directory/$osw_list_intermediary_detect_filename2");


    #***********************************************************************
    #
    #**********************************************************************
#    if($omatt_usecat eq "t" or $omatt_usecat eq "T")
#    {
#        if (not  -e $catfile)
#        {
#            my $command = "omcat set=$out_directory/$osw_list_intermediary_detect_filename outset=$catfile";
#            system( $command );
#        } 
#
#    }


    #***********************************************
    # omatt
    #***********************************************
    &HighlightedMessage("*", "omatt");

    
    $arg_list ="set=$detectorCoordImageFileName" .
        
#" set=$IntermediateImage3FileName" .
           #"set=$detectorCoordImageFileName" .
                " sourcelistset=$out_directory/$osw_list_intermediary_detect_filename" . 
		" ppsoswset=$skyCoordImageFileName" .
                " catfile=$catfile" .
                " usecat=$omatt_usecat" .
                " maxradecerr=$maxradecerr" .
                " maxrmsres=$maxrmsres" .
		" rotateimage=$omatt_rotateimage";
  
   
   


#    &Message("omatt $arg_list");
    system("omatt $arg_list") && SAS::fatal("Task Failure","omatt failed.");
    
    #**********************************************************
    # 12.6.2001
    # Add the file to the list if it is not the high-resolution
    # central image
    #**********************************************************
                  


    if (&IsHighResCentralWindow($osw_list_intermediary_detect_filename) == 0)
    {                   
	my $name = $out_directory . '/' . $osw_list_intermediary_detect_filename;
        push(@osw_file_list, $out_directory . '/' . $osw_list_intermediary_detect_filename);
	if ($skyCoordImageFileName =~ "SIMAGE00")
        {
            return;           
	}
      
        push(@Simage2000Files, $skyCoordImageFileName);
              
 
    }
       

}
#*********************************************
# GetFlatFieldImage
#*********************************************
sub GetFlatFieldImage($)
{
    my ($orb_key, $obs_key) = @_;   
  

    
    &HighlightedMessage("*", "omflatgen");
      
    $flatFieldFileName = "P$orb_key" . $obs_key . "X000FLAFLD0000.FIT";
        
    $flatFieldFileName = $out_directory . "/$flatFieldFileName";
    my $arg_list = "outset=$flatFieldFileName";

    my $message = "omflatgen $arg_list";
#    &Message($message);
    #********************************
    # Create an artificial one
    #********************************

    system("omflatgen $arg_list") && SAS::fatal("Task Failure", "omflatgen");


#    my $flat = "/opt/data/ssco2/odf/flatfield/pix2pix2006.fits";
#    if(-e $flat)
#    {
#        my $command = "cp $flat $flatFieldFileName";
#        print("$command\n");
#        system( $command );

#    }
#    else
#    {
#	die("file $flat not found\n");
#    }
    $in_orbit_flat_field_filename = $flatFieldFileName;  
      

  
    return $in_orbit_flat_field_filename;

}
#****************************************************
# Subroutine ProcessExposure
#****************************************************
sub ProcessExposure($)
{
    my ($orbitNo, $observationNo, $exposureNo, $exposureSchedule, $exposureType, $filter) = @_; 
    my ($type) = $ExposureType{$exposureNo};     # One of RUDI5, ENG2 or ENG4
 
    #***********************************************
    # Determine if RUDI5, ENG2 or ENG4 exposure
    #***********************************************

#    print("Processexposure line 946 exposuretype=$exposureType, filter=$filter\n");

 
    if ($exposureType eq "MISSING")
    {
     
        &HighlightedMessage("X", "Warning! - No image files found for exposure $exposureNo - ignoring exposure");
        return;    
          
    }
    my $text = "Processing exposure no $exposureNo (Imaging)";
    &HighlightedMessage("*", $text);


    my ($exposureKey) = $exposureType . $exposureNo;
    my ($i) = 0;
    my $directory;

    #**************************************************
    # If the exposure is an engineering-2 one combine
    # the four separate images into one.
    #**************************************************
    if ($exposureType eq "ENG2")
    {
        &CombineEng2Images($observationNo,  $exposureSchedule, $exposureNo);
        $directory = $out_directory;
    }
    else
    {
        $directory = $inp_directory;
    }

   
    #*******************************************************************
    # Check consistancy of odf.
    #
    # We need only check for the presence of the WDX and THX files here.
    #
    # We now have a true or dummy tracking history file name
    #*******************************************************************

    &CheckTHXFileExists($orbitNo, $observationNo, $exposureSchedule, $exposureNo);
   
    $window_data_filename = $inp_directory . "/". $orbitNo . "_" . $observationNo . "_OM" . $exposureSchedule . $exposureNo . "00WDX.FIT";

    if (&FileExists($window_data_filename) == 0)
    {    

        my $text = "Missing WDX file- skipping exposure";
        &HighlightedMessage("X", $text);
        return;     
    }     

    #**************************
    # Set pipeline variables.
    #**************************
  
    $tracking_history_plot_filename = $out_directory . "/P" . $observationNo . "OM" . $exposureSchedule . $exposureNo . "TSHPLT" . "000.ps";

     
    $tracking_history_plot_filename1 = $out_directory . "/P" . $observationNo . "OM" . $exposureSchedule . $exposureNo . "TSHPLT" . "000.PDF";

    $tracking_history_timeseries_filename = $out_directory . "/P" . $observationNo . "OM" . $exposureSchedule . $exposureNo . "TSTRTS000.FIT";

    #************************************
    # Tracking pipeline
    #************************************
    &HighlightedMessage("*", "Running Tracking pipeline");

    #********************************************
    #  omprep
    #********************************************
    &HighlightedMessage("*", "omprep");

    #********************************
    # The tracking-history file name
    #*******************************
    $trackingHistoryFileName = $out_directory . "/I" . $observationNo . "OM" . $exposureSchedule . $exposureNo . "TRHIS0000.FIT";


    my $arg_list = "set=$tracking_history_filename" .
                " nphset=$non_periodic_hk_filename" .
                " pehset=$periodic_hk_filename" .
                " wdxset=$window_data_filename" .
                " outset=$trackingHistoryFileName" .
                " modeset=3";                          

    system("omprep -v");
#    &Message("... omprep  $arg_list"); 
 
    system("omprep  $arg_list");

    my $exit_value = $? >> 8;   # 0 okay, 1 call fatal called, or some other disaster
    if ($exit_value == 1) # Failure in omprep 
    { 
        &Message("omprep has detected an error- this observation will not be processed");
        return;  
    }     
   

    #************************************************************
    # omdrifthst
    #************************************************************
    &HighlightedMessage("*", "omdrifthist");

    $arg_list = "set=$trackingHistoryFileName" .
                " plotfile=" .  $tracking_history_plot_filename;

#    &Message("omdrifthist $arg_list");

    system("omdrifthist $arg_list") && SAS::fatal("Task Failue","omdrifthist failed.");
	     

    #*******************************************************************
    # Only run ps2pdf if tracking-history file exists
    #*******************************************************************
    
    if (&FileExists($tracking_history_plot_filename) == 1) 
    {

        &HighlightedMessage("*", "ps2pdf");

        $arg_list = "$tracking_history_plot_filename $tracking_history_plot_filename1";
#        &Message("ps2pdf  $arg_list");
             
	  
        system("ps2pdf $arg_list") && SAS::fatal("Task Failure","ps2pdf failed.");
     
    }
    else
    {
        &HighlightedMessage("X", "Tracking-history-plot file name doesn't exist");
      
    }
    #************************************************************
    # omthconv
    #************************************************************
    &HighlightedMessage("*", "omthconv");

    $arg_list = "thxset=$trackingHistoryFileName" .
                " nphset=$non_periodic_hk_filename" .
                " outset=$tracking_history_timeseries_filename" ;

#    &Message("omthconv $arg_list");

    system("omthconv $arg_list") && SAS::fatal("Task Failure","omthconv failed.");
    

    #***************************************************
    # Loop over the windows (0 or 1) for RUDI5 images
    # (Low res & high res window)
    #***************************************************


    #**************************************************
    # Get a list of the IMI/E2I/E4I for this exposure
    #**************************************************
    my $extension = $ExposureExtension{$exposureNo};     # One of RUDI5, ENG2 or ENG4

    #***********************************************
    # Get a list of image files for this exposure
    #***********************************************

    my @FileList = &GetExposureFileList($directory, $exposureNo, $extension);

    #************************************************
    # If the exposure is a Rudi-5 one check that
    # there are 2 exposures (high and low resolution)
    #************************************************

    if ($exposureType eq "RUDI5")
    {
        &CheckRudi5Images($exposureNo, @FileList);
    } 

    #*********************************************
    # The imaging pipeline
    #*********************************************

    &HighlightedMessage("*", "Running Imaging pipeline");



    for (my $i = 0; $i < @FileList; $i++)
    {
                                     
        my $imageFileName = $FileList[$i];

         
        my $pos = index($imageFileName, $extension) - 24;
        my $name = substr($imageFileName, $pos, 32);    # Get rid of the directory name               
        my $windowNo = substr($name, 16, 2);            # Window no- 00 or 01
               
        &ProcessImageFile($imageFileName, $observationNo, $exposureSchedule, $exposureNo, $windowNo, $filter);
  
    }

}
#***************************
# Subroutine ProcessOrbit
#***************************
sub ProcessOrbit($)
{
    my ($orbitNo, @UserFilterList) = @_; 
       
    my $obs_list = $orb_list{$orbitNo};

    my $message;
 
    my @List = RetrieveList($orbitNo, %orb_list);

    #************************************************
    # Process observations for the orbit number
    #***********************************************
    for (my $i = 0; $i < @List; $i++)
    {
     
        my $exposureNo;
        my $oldExposureType; 
        my $exposureType;
 


	$obs_key = $List[$i];

        #*************************************************
        # Get the flat-field (create one if necessary)
        #*************************************************     
        $in_orbit_flat_field_filename = &GetFlatFieldImage($orbitNo, $obs_key); 
       
     
       
        &HighlightedMessage("*", "Observation $obs_key");
       
    
        #**************************************************
        # Process the data for each exposure of each filter
        #**************************************************
  
        @UserExposureList = &GetStringParameterList("exposures");


        if (@UserFilterList > 0 and @UserExposureList > 0)
        {
            &HighlightedMessage("X", "Error! - You can only supply either a list of filters or a list of exposures- aborted");
            die;    
        } 
        my $userSupplied = 0;
        my @List = 0;
        if (@UserFilterList > 0)
        {
            #*****************************************************
            # Check that the list is valid and modify accordingly
            #*****************************************************
            @List = &CheckUserFilterList(@UserFilterList);
             
        }
        elsif (@UserExposureList > 0)
        {
           

            #*****************************************************
            # Obtain the list of filters for the user-specified
            # exposures
            #*****************************************************
            @List = &CheckUserExposureList(@UserExposureList);



            $userSupplied = 1;
        }
        else
        {
            @List = &GetFilterList;  # List of OM filters in summary file
        }
      
        for (my $k = 0; $k < @List; $k++)
        {
	  


           #***********************************************
           # Stop processing image file if a grism image
           #***********************************************
  
       
            @Simage2000Files = () ;    # Stores a list of the image files for mosaicing    
    
	    my @List1 = &RetrieveList($List[$k], %ExposureFilterList);

            if ($userSupplied == 1)
            {
		
                my @temp = ();
                my $l = 0;
                for ($l = 0; $l < @UserExposureList; $l++)
                {
                    my $m = 0;
                    my $found = 0;
                    while ($m < @List1 and $found == 0)
                    {
                        if($UserExposureList[$l] eq $List1[$m])
                        {
                            $found = 1;
                            push(@temp, $UserExposureList[$l]);
                        }
                        $m++;
                    }
                 }
                 @List1 = @temp;
      
	     } 

            my $noExposures = @List1;
            my $message;
            if ($noExposures == 0) 
            {
                next;
            }


           if ( $List[$k] eq "GRISM1" or $List[$k] eq "GRISM2") 
           {
	          my $message = "Ignoring filter $List[$k]- please use the omgrism chain to reduce the data";
                  &HighlightedMessage("X", $message);
        
                  next;

           }
	 


	    if ($noExposures == 1)
            {
                $message = "Processing $List[$k] band image (exposure " . join(", ", @List1) . ")";   
            }
            else
            {    
                $message = "Processing $List[$k] band images (exposures " . join(", ", @List1) . ")";   
            }
	    &HighlightedMessage("*", $message);
             
                    
            my $start = time();
 
         
            for (my $i = 0; $i < $noExposures; $i++) 
            {


                $exposureNo = $List1[$i];
                $exposureType = &FindExposureType($exposureNo);
   



	       
                if ($exposureType ne "MISSING")
                {
                    $oldExposureType = $exposureType;
		  
                } 
                my $exposureSchedule = $ExposureScheduleList{$exposureNo};
         
                my $exposure_key = $exposureSchedule . $exposureNo;
         
           
                &ProcessExposure($orbitNo, $obs_key, $exposureNo, $exposureSchedule, $exposureType,  $List[$k]);
	       

            }
            $exposureType = $oldExposureType;

            #****************************************
            # Mosaic the images for thsi filter
            #****************************************    
 	    my $num =  @Simage2000Files;
	  
            if ($num > 0)
            {
                 
                &MosaicImages($exposureType, $List[$k], @Simage2000Files);    

            }


#            else
#            {
#                if ($num == 1)
#                {
#                    my $zero = 0;
#                    &HighlightedMessage("*", "Warning- there is only one sky-coord image ($Simage2000Files[$zero]" . "- ommosaic will not run");
#                }
#            } 
            my $time = time() - $start;
            if ($time > 1)
            {
                my $message = "Finished processing images for filter $List[$k] - time taken= $time  seconds";
	        &HighlightedMessage("*", $message);
            }       
        } 

        #***********************************************
        # Check that there is at least one source-list
        #***********************************************
        my $numSourceLists = @osw_file_list;

        if ($numSourceLists > 0) 
        {  


            #**********************************
            # Set remaining pipeline variables.
            #**********************************
 

            $observation_source_list_pps_product_filename = "P$obs_key" ."OMCOMBOBSMLI0000.FIT";

            #**********************************************
            # omsrclistcomb.
            #********************************************** 
            &HighlightedMessage("*", "omsrclistcomb");

            my $arg_list = "sourcelistsets=\"" . join(" ", @osw_file_list) . "\"" .   
		           " nsigma=$omsrclistcomb_nsigma" .
                           " usecat=$omsrclistcomb_usecat" .
                           " catfile=$catfile" .
                           " maxradecerr=$maxradecerr" .
                           " maxrmsres=$maxrmsres" .
                           " outset=$out_directory/$observation_source_list_pps_product_filename";

#            &Message("omsrclistcomb $arg_list");

            system("omsrclistcomb $arg_list") && SAS::fatal("Task Failure","omsrclistcomb failed.");

        }
        else
        {
            &HighlightedMessage("Warning!- There are no source-list files for omsrclistcomb to process");
        }
    }

}
#*******************************
# sub SetUpDirectoryPaths
#*******************************
sub SetUpDirectoryPaths
{

    $inp_directory = stringParameter("inpdirectory");


 



 ############ VNY 27.01.2004
    my $directory;
    if (length($inp_directory) < 2 || length($inp_directory) >255)
    {
        #*************************************************
        # If the SAS_ODF environment variable has been set
        # then set the input directory to it- otherwise
        # set the input directory to the current directory
        #*************************************************
        my $len = 0;
        my $input = " ";
        if (exists($ENV{SAS_ODF}))
        {
            $len = length($ENV{SAS_ODF});
            $input = $ENV{SAS_ODF};
        }      

        if ($len < 3)
        {
            $inp_directory = &GetCurrentDirectory;
            $ENV{SAS_ODF} = $inp_directory;
            my @SASFiles = <$inp_directory/*.SAS>;        
            $SASFile = shift @SASFiles;
        }
        else
        {     
            #*************************************************
            # If SAS_ODF is set to a .SAS file then need
            # to get the ODF directory from it
            #************************************************

            my $pos = index($input, "SAS");
          
             
            if ($pos > 0) 
            {
	        my $length = length($input);
                $directory = substr($input, 1, $length - 4); 
                $SASFile = $input;
		$inp_directory = &GetODFDirectory;   
            }
            else
            {
                $inp_directory = $input;
             
                my @SASFiles = <$inp_directory/*.SAS>;        
        
 
                $SASFile = shift @SASFiles;
                my $string =  "$inp_directory/*.FIT";
                     
            }
        }

    }
    else
   {
       my @SASFiles = <$inp_directory/*.SAS>;        

       $SASFile = shift @SASFiles;
       $ENV{SAS_ODF} = $inp_directory;
   }   
   
    $out_directory = stringParameter("outdirectory");
########### VNY 27.01.2004    
    if (length($out_directory) < 2 || length($out_directory) > 255)
    {
	$out_directory = &GetCurrentDirectory;
      
    }
    if (FileExists($out_directory) == 0) 
    { 
#        SAS::fatal("Task Failure","$out_directory doesn't exist");
    die("Task Failure- $out_directory doesn't exist");

    }
############ VNY 27.01.2004 
    if (length($inp_directory) < 2 || length($inp_directory) > 255)
    {
	$inp_directory = &GetCurrentDirectory; 
    }
    if (FileExists($inp_directory) == 0) 
    { 
#        SAS::fatal("Task Failure","$inp_directory doesn't exist");
        die("Task Failure- $inp_directory doesn't exist");
    }

  
}
#**********************************
# sub CheckForHouseKeepingFiles
#**********************************
sub CheckForHouseKeepingFiles
{
    if($periodic_file_flag == 0)
    {
        &Message("The ODF does NOT contain a Periodic Housekeeping file. Please insert into the ODF. ");
        SAS::fatal("Task Failure","Incomplete ODF.");
    }

    if($non_periodic_file_flag == 0)
    {
        &Message("The ODF does NOT contain a Non Periodic Housekeeping file. Please insert into the ODF.");
        SAS::fatal("Task Failure","incomplete ODF.");
    }
}

#**************************************
# Subroutine InitialiseLists
#**************************************
sub InitialiseLists
{
   
#    print("initilaiselists line 1509\n");

    &FillExposureNoList;           # Stores a list of exposures from the summary file. 

   
}
#**************************************
# Subroutine Message
#**************************************
sub Message($)
{ 
    my ($message) = @_;  

#    my $length = length($message);

#    if ($length <= 72)
#    { 

#    for (my $i = 0; $i < $maxLength; $i++)
#    { 
#        $text = $text . $character;
#    }
   




        if ($redirect == 5) 
        {
            print("$message \n");
        }
        else  
        {
            SAS::message($SAS::AppMsg, $SAS::SparseMsg, $message);



        }
#    }
#    else
#    {
#       my $i = 0;
#       my $j = 0;
#       do {
       
#	    $j += 72;
#            my $message1 = substr($message, $i, 72);
#            if ($redirect == 5) 
#            {
#                print("$message1 \n");
#            }
#            else  
#            {
#                SAS::message($SAS::AppMsg, $SAS::SparseMsg, $message1);



#            }
#            $i += 1;
#        } until ($j >= $length);  

#    }

}
#**************************************
# Subroutine HighlightedMessage
# Outputs lines of text
#**************************************
sub HighlightedMessage($)
{ 
    
    my ($character, @Messages) = @_;  
   
    my $noLines = @Messages;
    my $maxLength = length($Messages[0]);
 
    if ($noLines > 1)
    {
        for (my $i = 1; $i < $noLines; $i++)
        {
            if (length($Messages[$i]) > $maxLength)
	    {
                $maxLength = length($Messages[$i]);
            }
        } 
    }
######## VNY 27.01.2004
    if ($maxLength > 80)
    {
	$maxLength=80;
    }
########
    my $text = "";
   
    for (my $i = 0; $i < $maxLength; $i++)
    { 
        $text = $text . $character;
    }
######### VNY 27.01.2004
#    &Message("\n$text");
    &Message("$text");
########    
    for (my $i = 0; $i < $noLines; $i++)
    { 
       
        &Message($Messages[$i]);
    }

    &Message($text);
   

}

#***************************************
# CreateLogFile
#**************************************
sub CreateLogFile
{
   
    $redirect = 5;
   
    open (STDOUT, ">omichain_log"); # and die("Couldn't open file");;
 
    
 
    open (STDERR, ">&STDOUT");
 
 
    select (STDERR);
  
 
}
#***********************************
# Subroutine GetCurrentDirectory
#**********************************
sub GetCurrentDirectory
{
    my $command = "pwd";
    my $curdir = `$command`;

    my $leng = length($curdir);
    substr($curdir, $leng - 1) = "";    
    return $curdir;
}


#******************************
# Subroutine GetODFDirectory
# Retrieves the directory-path
# of the ODF data files.
#******************************
sub GetODFDirectory($)
{
     
   

    open (SASFILE, $SASFile) || SAS::fatal("OMICHAIN","Unable to open SAS SUMMARY FILE $SASFile"); 
 

    my $line = <SASFILE>;
        
  
    while ($line = <SASFILE>)
    {
        
        if ($line =~ "PATH")
        {        

            my $length = length($line);
            my $directory = substr($line, 5, $length - 6); 
            #if ($directory eq "." or $directory eq "./")
            #{
            #    my $length = length($SASFile);
            #    $directory = substr($SASFile, 0, $length - 32);
            #}
            close SASFILE;
            return $directory; 
             
        }
    }
    close SASFILE;


}

#************************************
# Sub GetOmmodmapParameters
#************************************
sub GetOmmodmapParameters
{
    $ommodmap_nsig = intParameter("ommodmapnsig");
    $ommodmap_nbox = intParameter("ommodmapnbox"); 
######## VNY 27.01.2004
    if ($ommodmap_nsig > 10 || $ommodmap_nsig < 1)
    {
	$ommodmap_nsig=3;
    }
    if ($ommodmap_nbox > 2048 || $ommodmap_nbox < 1)
    {
	$ommodmap_nbox=16;
    }
}
#************************************
# Sub GetOmdetectParameters
#************************************
sub GetOmdetectParameters
{
  
    $omdetect_nsigma = realParameter("omdetectnsigma");
    $omdetect_minsignificance = realParameter("omdetectminsignificance");
#    print("omdetectminsignificance= $omdetect_minsignificance\n");

    $omdetect_detectextended = booleanParameter("omdetectdetectextended");
######### VNY 27.01.2004
    if ($omdetect_nsigma > 20.0 || $omdetect_nsigma < 1.0 )
    {
	$omdetect_nsigma=2.0;
    }
    if ($omdetect_minsignificance < 20.0 and $omdetect_minsignificance > 0.0 )
    {
	$omdetect_minsignificance=$omdetect_minsignificance * 1.0;
    }
    else
    {
	$omdetect_minsignificance = 0.0;
    }
    if ($omdetect_detectextended eq 0 or $omdetect_detectextended eq "no" 
	or $omdetect_detectextended eq "NO" )
    {
	$omdetect_detectextended="F";
    }
    else
    {
	$omdetect_detectextended="T";
    }

}
#************************************
# Sub GetOmsrclistcombParameters
#************************************
sub GetOmsrclistcombParameters
{
   
    $omsrclistcomb_nsigma = realParameter("omsrclistcombnsigma");
    $omsrclistcomb_usecat = &GetBooleanParameter("omsrclistcombusecat");

 
    if ($omsrclistcomb_nsigma >10.0 || $omsrclistcomb_nsigma < 1.0)
    {
	$omsrclistcomb_nsigma = 3.0;
    }

  
}
#************************************
# Sub GetOmmosaicParameters
#************************************
sub GetOmmosaicParameters
{
  
    $ommosaic_exposure = realParameter("ommosaicexposure");
    $ommosaic_exposuremap = &GetBooleanParameter("ommosaicexposuremap");
    if ($ommosaic_exposure > 100000.0 || $ommosaic_exposure < 10.0)
    {
	$ommosaic_exposure =1000.0;
    } 
}

#************************************
# Sub GetOmattParameters
#************************************
sub GetOmattParameters
{    
    $omatt_usecat = &GetBooleanParameter("usecat");
    $catfile = stringParameter("catfile");
   if (length($catfile) < 2 || length($catfile) >255)
   {
       $catfile=$inp_directory/"usnocat.fit";

   }
 
    $omatt_rotateimage =  &GetBooleanParameter("omattrotateimage");


    $maxradecerr =  realParameter("maxradecerr");
    $maxrmsres = realParameter("maxrmsres");

}
#**************************************
# GetBooleanParameter
#**************************************
sub GetBooleanParameter($)
{
    my ($name) = @_;
    my ($parameter) = booleanParameter($name);
   
    my ($converted) = "F";
    if ($parameter eq "1" or $parameter eq "yes" or $parameter eq "YES")
    {
        $converted = "T";
    }
  
    return $converted;
}
#**************************************
# Sub GetFilterList
# Returns a list of the OM filters
# in the SAS summary file
#**************************************
sub GetFilterList
{

   open (SASFILE, $SASFile) || SAS::fatal("OMICHAIN","Unable to open SAS SUMMARY FILE $SASFile"); 
 
    #*************************************************
    # Skip past the EPIC and RGS stuff
    #*************************************************
   
  
    my $line = <SASFILE>;
       
    until (substr($line, 0, 2) eq "OM")
    {

    $line = <SASFILE>;

                                  
    } 

    @FilterList = (); 

    while ($line = <SASFILE>)

    {
        
        if ($line =~ "/ Exposure Identifier")
        {                 


	   my $filterName = &GetNextSASFilter;
	  
           if ($filterName eq "")
           {
               last;
           }       
           
            
            #*********************************************
            # Add the filter to the list
            #*********************************************
            my $numFilters = @FilterList;
            my $exists = 0;
       
        
	    my $i = -1;
            while (++$i < $numFilters and $exists == 0) 
            {
	   
                if ($filterName eq $FilterList[$i])
                {
                
                    $exists = 1;
               

                }

            }
	    
            if ($exists == 0)  # Found a different filter
            {
 
                if (&FilterOkay($filterName) == 0)
                {
                    push(@FilterList, $filterName);
                }
	    }
        }
    }


   close SASFILE;
    
    return @FilterList;
    
}
#************************************************
# Sub GetFilterFileList
# Returns a list of files for a
# given filter which contain
# the given expression
# Usage :: @List = @GetFilterList("V", "*IMI*");
#************************************************
sub GetFilterFileList($)
{
    my ($filter, $string) = @_;  
    my($i, $j) = 0;
    
    open (SASFILE, $SASFile) || SAS::fatal("OMICHAIN","Unable to open SAS SUMMARY FILE $SASFile"); 
 
    #*************************************************
    # Skip past the EPIC and RGS stuff
    #*************************************************

 
    my $line = <SASFILE>;
       
    until (substr($line, 0, 2) eq "OM")
    {
        $line = <SASFILE>;
                                  
    } 
    my @List =();           # Initialise the file list    
 
    while ($line = <SASFILE>)
    {
        
        if ($line =~ "FILTER =")
        {
         
            my $pos = index($line, "//");

            my $filterName = substr($line, 9, $pos - 10); # Get rid of the characters after the end of the filter
	    
            if ($filterName eq "White")
            {
                $filterName = "WHITE";
	    }
            if ($filterName eq "Grism1")
            {
                $filterName = "GRISM1";
	    }

            if ($filterName eq "Grism2")
            {
                $filterName = "GRISM2";
	      
	    }
             
            if ($filterName eq "Blocked")
            {
                $filterName = "BLOCKED";
	
            }
     

            if ($filterName eq $filter)
            {  
                until ($line =~ "/ Exposure Identifier")
               
                {
	            $line = <SASFILE>;
           
                } 
                 
                my $exposureNo = substr($line, 0, 3);
                push(@List, $exposureNo);         
                     
            }  
        }
    }
   
    close SASFILE;
    #******************************************************
    # Select all the files containing the specified string    
    #******************************************************
    my @FileList = <$inp_directory/$string>;        

      
    @FilterFileList =();           # Initialise the file list    
 
    for (my $i = 0; $i < @FileList; $i++)
    { 
       
        for (my $j = 0; $j < @List; $j++)
        {
            if ($FileList[$i] =~ $List[$j])
            {
                push(@FilterFileList, $FileList[$i]);         
            } 
        }
    }
    return @FilterFileList;  
    
}
#*****************************
# Subroutine Display
# Displays the given list
#****************************
sub Display($)
{
    my ($title, @List) = @_;
   
#    print("DISPLAYING LIST $title \n"); 
 
  
    for (my $i = 0; $i < @List; $i++)
    {
        print("$i) $List[$i] \n");
    }
}
#*******************************************
# GetNextSASFilter
#******************************************
#sub GetNextSASFilter($)
sub GetNextSASFilter

{
    my $line = <SASFILE>;
 
    
    until($line =~ "FILTER =" or $line eq "")
    {
        $line = <SASFILE>;
       
    } 
    if ($line eq "")
    {
      
        return $line;

    }
            
    my $pos = index($line, "//");

    my $filterName = substr($line, 9, $pos - 10); # Get rid of the characters after the end of the filter

    if ($filterName eq "White")
    {
        $filterName = "WHITE";
    }


    if ($filterName eq "Grism1")
    {
        $filterName = "GRISM1";
    }

    if ($filterName eq "Grism2")
    {
        $filterName = "GRISM2";
	
    }
             
    if ($filterName eq "Blocked")
    {
        $filterName = "BLOCKED";
	
    }
             
   
    return $filterName;
}
#************************************
# GetExposureNoListForFilter      
# Returns a list of the exposures
# made for the given filter
#************************************
sub GetExposureNoListForFilter($)
{
    my ($filter, $exposureNo) = @_;  
    
       
    my @ExposureNoFilterList = ();

    my @ExposureNoList = RetrieveList($filter, %ExposureFilterList);


    return  @ExposureNoList;  
}    
#*******************************************
# Sub FindExposureType
# Determines if the given exposure number
# is for a RUDI5 sequence, or ENG2
#*******************************************
sub FindExposureType($)
{
    my ($exposureNo) = @_;  
 
    my(@ImageFiles) = <$inp_directory/*IMI.FIT*>;           

    my $num;
    my $numFiles = @ImageFiles;
      
    @ExposureImageList = ();           # Initialise the file list    
     
    my $type;

    for (my $j = 0; $j < $numFiles; $j++)
    {
               
        my $fileName = $ImageFiles[$j];       
        my $pos = index($fileName, "IMI") - 24;

        my $string = substr($fileName, $pos, 32);    # Get rid of the directory name

        my $text = substr($string, 19, 3);           # The exposure number
    
        if ($text eq $exposureNo)
        {
            push(@ExposureImageList, $fileName);         
          
        }            
              
                  
    }


    my $numIMIFiles = @ExposureImageList;
   
    $num = 0;
    
    
    
    if ($numIMIFiles > 0)
    {

        for (my $j = 0; $j < $numIMIFiles; $j++)
        {
        
            my $fileName = $ExposureImageList[$j];       
        
            my $pos = index($fileName, "IMI") - 24;
            my $name = substr($fileName, $pos, 32);    # Get rid of the directory name               
            my $pattern = "0" .  $j . "IMI";  
            $pos = index($name, $pattern);
            if ($pos > 0) 
            {
                $num++;
            }
        }
#	print("numimfiles = $numIMIFiles num=$num\n");
#	die();

# MODIFICATION ON 2nd NOVEMBER


   if ($num == 4 or $num == 3)
     
#        if ($num == 4)
        {
            $type = "ENG2";
            $ExposureType{$exposureNo} = "ENG2";
	    $ExposureExtension{$exposureNo} = "*ENG2.FIT*";
	
        }
        else
        {
            $type = "RUDI5";
            $ExposureType{$exposureNo} = "RUDI5";
	    $ExposureExtension{$exposureNo} = "*IMI.FIT*";
	
        }
    }
    else
    {


        my(@ImageFiles) = <$inp_directory/*E4I.FIT*>;           
        $numFiles = @ImageFiles;
   

        for (my $j = 0; $j < $numFiles; $j++)
        {
               
            my $fileName = $ImageFiles[$j];       
            my $pos = index($fileName, "E4I") - 24;

            my $string = substr($fileName, $pos, 32);    # Get rid of the directory name

            my $text = substr($string, 19, 3);           # The exposure number
    
            if ($text eq $exposureNo)
            {
                push(@ExposureImageList, $fileName);         
          
            }            
              
                  
        }

        my $numFiles = @ExposureImageList;
      
  
        if ($numFiles == 0 )
        {
	   
            $type = "MISSING";
        }
        else
        {
	    $type = "ENG4";
            $ExposureType{$exposureNo} = "ENG4";
            $ExposureExtension{$exposureNo} = "*E4I.FIT*";
	
        }

    }
    return $type;
    
}
#******************************************************
# Sub MosaicImage
# Mosaics the given list of
# images
#*****************************************************
sub MosaicImages($)
{

 
    if ($omatt_rotateimage eq 0) 
    {

        return;
    }

  
  



    my($exposureType, $filter, @ImageList) = @_;


 

    my $text = "ommosaic- mosaicing images for filter $filter";
    &HighlightedMessage("*", $text);
    
    my $string; 

    if ($exposureType eq "RUDI5")
    { 
        $string = "RSIMAG";
    }
    elsif($exposureType eq "ENG2")
    {
        $string = "LSIMAG";
    }
    elsif($exposureType eq "ENG4")
    {
        $string = "HSIMAG";
    }
    elsif($exposureType eq "MISSING")
    {
        return;
    }
    else

    {
        die ("UNIDENTIFIED EXPOSURE TYPE ($exposureType)");


    }
            




    my $filterCode = &GetFilterCode($filter);

    my $position = index($ImageList[0], "OMS") - 11;
    
   
    my $outputFileName = $out_directory . "/" . substr($ImageList[0], $position, 14) . "000$string" . $filterCode . ".FIT";
     

   
    my $arg_list = "imagesets=\"" . join(" ", @ImageList) . "\" mosaicedset=$outputFileName" . 
                   " exposuremap=$ommosaic_exposuremap " . "exposure=$ommosaic_exposure";

#    &Message("ommosaic $arg_list");



    system("ommosaic $arg_list") && SAS::fatal("Task Failure", "ommosaic");

}
#************************************************
# Sub GetFilterFileList
# Returns a list of files for a
# given filter which contain
# the given expression
# Usage :: @List = @GetFilterList("V", "*IMI*");
#************************************************
sub GetExposureFileList($)
{

    my ($directory, $exposureNo, $string) = @_;  
  
    #******************************************************
    # Select all the files containing the specified string    
    #******************************************************
    my(@FileList) = <$directory/$string>;        
    
    my @ExposureFileList =();           # Initialise the file list    
 
    for (my $i = 0; $i < @FileList; $i++)
    { 
        my $position = length($FileList[$i]) - 12; 
        if (substr($FileList[$i], $position, 3) =~ $exposureNo)
  
        {
	    
            push(@ExposureFileList, $FileList[$i]);         
        } 
    }
    
    return @ExposureFileList;  
    
}
#***************************************
# Subroutine CheckRudi5Images
#***************************************
sub CheckRudi5Images($)
{
    my ($exposureNo, @FileList) = @_; 

    my $number = @FileList;

    if ($number ne 2)
    {

        my $text = "Warning! - Missing image file";
        &HighlightedMessage("X", $text);
    }
}
#******************************************
# Subroutine CombineEng2Images
# Combines the 4 separate ENG2 image files
# into a single image
#******************************************    
sub CombineEng2Images($)
{
 
    my ($observationNo, $exposureSchedule, $exposureNo) = @_;
            
    my($fileName) = $ExposureImageList[0];       
    my($pos) = index($fileName, "IMI") - 24;
    my($name) = substr($fileName, $pos, 32);    # Get rid of the directory name                          

    #*****************************************************************
    # The new image file is temporarily stored in the input directory 
    #*****************************************************************  

    my $outputFileName = $out_directory . "/I$observationNo" . "OM$exposureSchedule$exposureNo" . "00E2I.FIT";
  
  
    

    &HighlightedMessage("*", "omcomb");
    my $arg_list = "imagesets=\"" . join(" ", @ExposureImageList) . "\"" . " outset=$outputFileName";

#    &Message("omcomb $arg_list");
              
    system("omcomb $arg_list") && SAS::fatal("Task Failure", "omcomb");
    #***********************************************
    # Check to see if omcomb ran successfully
    #***********************************************
    my $exit_value = $? >> 8;   # 0 okay, 1 call fatal called, or some other disaster
   
    if ($exit_value == 1)  
    { 
       

	$ExposureType{$exposureNo} = "RUDI5";
	$ExposureExtension{$exposureNo} = "*IMI.FIT*";
	                
     
        &Message("The test in omcomb have failed - This exposure ($exposureNo) will be treated as normal imaging data");
        
    }      
    else
    {

                          
        $ExposureType{$exposureNo} = "ENG2";
	$ExposureExtension{$exposureNo} = "*E2I.FIT*";
		
        push(@E2IMIList, $outputFileName);   # A list of the combined image files           
        	
    }  
}
#************************************
# Subroutine FilterOkay
#************************************
sub FilterOkay($)
{
    my ($filter) = @_;
    
    my (@BadFilter) = ("BLOCKED", "Blocked");
    
    for (my $i = 0; $i < @BadFilter; $i++)
    {
	
        if ($filter eq $BadFilter[$i])
        {
            my $message = "Warning! - Cannot process filter $filter. The image files for this filter will not be processed";
            &HighlightedMessage("X", $message);
         
            return 1;
        }
    }
    
    return 0;
}
#***************************
# Subroutine GetFileList
#***************************
sub GetFileList($)
{
    my (@Pattern) = @_;
    my @Filelist = ();
    

    
    for (my $i = 0; $i < @Pattern; $i++)
    {
	
        my (@List) = <$Pattern[$i]>;
        for (my $j = 0; $j < @List; $j++)
        {
            push(@FileList, $List[$j]);
        }
    }
    &Display("test", @FileList);
    return @FileList;
}         
#*****************************************
# Checks to see if the given file exists
# Returns 1 if it does, 0 otherwise.
#
#*****************************************
sub FileExists($)
{

    my ($file) = @_;
  
    if (-e $file)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}
#***************************
# GetStringParameterList
# Gets the list of strings
# in the command line for a
# specified parameter name
#***************************
sub GetStringParameterList($)
{
  
    my($parameter) = @_;
    my $num = 0;
    $num = parameterCount($parameter);
    my (@ParameterList) = ();
   
   
    if ($num eq 0)
    {
        return @ParameterList;
    }
  
  
   for (my $i = 0; $i < $num; $i++)
   {
        $ParameterList[$i] = stringParameter($parameter, $i);

   }
   return @ParameterList; 
   
}
#************************************************************
# Sub CheckUserFilterList
# Checks that the user-supplied list of filters exist in 
# the SAS summary file and returns a new list (if necessary)
#************************************************************  
sub CheckUserFilterList($)
{
    my(@UserFilterList) = @_;
    my(@List) = &GetFilterList;  # List of OM filters in summary file
      
   
    my @ModifiedList = ();

    for (my $i = 0; $i < @UserFilterList; $i++)
    {
        my $found = 0;
        my $j = 0;
        while ($j lt @List and $found == 0)
        {
            if ($List[$j] eq $UserFilterList[$i])
            {  
                push(@ModifiedList,  $UserFilterList[$i]);
                $found = 1;
            }
            $j++;
        }
        if ($found == 0)
        {
            my $message = "Warning! - Filter $UserFilterList[$i] does not exist in SAS Summary file- ignoring filter";
            &HighlightedMessage("X", $message); 
        }
    }
    if (@ModifiedList == 0)
    {
        @OMFilterList = &GetFilterList; # The list of OM filters from the SAS summary file
     
        my $message = "Error! - No valid filters to process (Valid filters are ". join(", ", @OMFilterList) .")- Aborting";
    
        &HighlightedMessage("X", $message);
       
        die;
    }
    return @ModifiedList;
}
#*******************************
# Sub CheckUserExposureList
# Checks that the command-line 
# list of exposures is valid
# (ie the filters are present)
# and get a list of the filters
# for the exposure list
#*******************************
sub CheckUserExposureList($)
{

    @OMFilterList = &GetFilterList; # The list of OM filters from the SAS summary file


    

    #**************************************************
    # Loop through the user-supplied list of exposures
    # and produce a filter list
    #**************************************************


    for (my $i = 0; $i < @UserExposureList; $i++)
    {
        my $found = 0;
        my $j = 0;

        #******************************************************************
        # Loop through the filter list obtained from the SAS summary file
        #******************************************************************
        while ($j < @OMFilterList and $found == 0)
        {

 
            #****************************************************
            # Get a list of the exposures for this filter
            #****************************************************
         
            my (@ExposureNoList) = &GetExposureNoListForFilter($OMFilterList[$j]);
            my $k = 0;

            while ($k < @ExposureNoList and $found == 0)
            {
               
                push(@ExposureList, $ExposureNoList[$k]); 
                if ($ExposureNoList[$k] eq $UserExposureList[$i])
                {
                    my $filterExists = 0;

		    my $l = 0;
                    while($l < @FilterList and $filterExists == 0)
                    {
                       
                        if ($FilterList[$l] eq  $OMFilterList[$j])
                        {
                      
                            $filterExists = 1;
                            last;
                        }
                        $l++;
                    }
                    if ($filterExists == 0) 
                    {
                        push(@FilterList, $OMFilterList[$j]);
                    } 
                    $found = 1;
                  
                }
                $k++;
            }
            $j++;
	}
        if ($found == 0) 
        {
            &HighlightedMessage("X", "Warning! - Invalid exposure no ($UserExposureList[$i])");
        }  
           
    }
  
    if (@FilterList == 0)
    {

        my @Message = ();
        push (@Message, "Error!- Invalid list of exposures- aborted");
        for (my $i = 0; $i < @OMFilterList; $i++ )
        {
            my (@ExposureNoList) = &GetExposureNoListForFilter($OMFilterList[$i]);
            my $text = "Valid exposures for filter $OMFilterList[$i] are "; 
      
            $text = $text . join(", ", @ExposureNoList);   
            push(@Message, $text);
	}
        &HighlightedMessage("X", @Message);
    }
 
  
    return @FilterList;
}                 

#***************************************************
# Purpose:- Given an hash list and a key, stores
# the given value
#
#**************************************************
sub Store
{
    my ($key, $value, %self) = @_;
   
    my %List = %self;     

    push @{$List{$key}}, $value;
   
 
    
    return %List;
}
#***************************************************
# Purpose:- Given an hash list and a key, returns
# an array containing the values for that key 
#
#**************************************************
sub RetrieveList
{
    my ($key, %self) = @_;
    
    my @List = ();
    my $n =  @{$self{$key}};
 
    for (my $i = 0; $i < $n; $i++)
    {
    
        my $value =  $self{$key}[$i];

        push(@List, $value);

    }

    return @List;
}
#******************************************************
# ExamineSASFile
# Obtains a list of the exposures in the summary file
#******************************************************  
sub ExamineSASFile
{

    my %ObservationIdentifier = ();
    my %RevolutionIdentifier = ();
    open (SASFILE, $SASFile) || SAS::fatal("OMICHAIN","Unable to open SAS SUMMARY FILE $SASFile"); 
  

   
    #*************************************************
    # Skip past the EPIC and RGS stuff
    #*************************************************
 
    $periodic_file_flag = 0;
    $non_periodic_file_flag = 0;
    my $observationId = "blank";
    my $revolutionId = "blank";
 
    my $line;
    

    while ($line = <SASFILE>)
    {
#      print("Line 2670 line=$line \n");        
       if ($line =~ "/ ODF Constituent" and $line =~ "OM")    # An ODF file

#   if ($line =~ "file" and $line =~ "OM")    # An ODF file


       {
            my $file = substr($line, 0, 31);
#print("Line 2674 file=$file \n");                        
            if ($file =~ "NPH")    # An NPH file
            {
                $non_periodic_file_flag = 1;
                $non_periodic_hk_filename = $inp_directory . "/" . $file;

            } 
            if ($file =~ "PEH")    # An PEH file
            {
                $periodic_file_flag = 1;
                $periodic_hk_filename =  $inp_directory ."/" . $file;


            } 
       }
       if ($line =~ "Observation Identifier") 

#   if ($line =~ "Observation Identifier" or $line =~ "Exposure Id") 


       {
           $observationId = substr($line, 0, 10);
           $ObservationIdentifier{$observationId} = $observationId;

#	   print("Observation identifier =$observationId\n");


#       if ($revolutionId ne "blank" and $observationId ne "blank")
#           {
       
#	       print("Storing in orb list $revolutionId, $observationId \n");
#               %orb_list = Store($revolutionId, $observationId, %orb_list);
#           }







       }


       if ($line =~ "Revolution Identifier")

#    if ($line =~ "Revolution Identifier" or $line =~ "Revolution number")

       {

           $revolutionId = substr($line, 0, 4);

#	   print("Revolution identifier =$revolutionId\n");


           $RevolutionIdentifier{$revolutionId} = $name;
           if ($revolutionId ne "blank" and $observationId ne "blank")
           {
       
#	       print("Storing in orb list $revolutionId, $observationId \n");
               %orb_list = Store($revolutionId, $observationId, %orb_list);
           }

       }
                                
    } 
    close SASFILE;

}











